/* Copyright (C) 2018 by Intevation GmbH <info@intevation.de>
 *
 * This file is Free Software under the GNU GPL (v>=2)
 * and comes with ABSOLUTELY NO WARRANTY!
 * See LICENSE.txt for details.
 */

#include "gpgolconfigpage.h"
#include "w32-gettext.h"
#include "w32-util.h"

#include <QDebug>
#include <QGroupBox>
#include <QLabel>
#include <QCheckBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QToolTip>
#include <QPushButton>
#include <QFileDialog>
#include <QComboBox>
#include <QLineEdit>
#include <QStandardPaths>
#include <QDir>

/* See gpgol/src/debug.h */
#define DBG_OOM            (1<<1)
#define DBG_MEMORY         (1<<2)
#define DBG_TRACE          (1<<3)
#define DBG_DATA           (1<<4)

/*
class ExplainingChkBox: public QWidget
{
    Q_OBJECT
public:
    explicit ExplainingChkBox(const QString &text, const QString &explanation):
        mChkBox(new QCheckBox(text)),
        mExplanation(explanation)
    {
        auto hBox = new QHBoxLayout(this);

        hBox->addWidget(mChkBox);

        auto infoBtn = new QPushButton;
        infoBtn->setIcon(QIcon::fromTheme("help-contextual"));
        hBox->addWidget(infoBtn);
        hBox->addStretch(1);

        connect(infoBtn, &QPushButton::clicked, this, [this, infoBtn] () {
                QToolTip::showText(infoBtn->mapToGlobal(QPoint()), mExplanation, infoBtn);
            });
    }

    void setChecked(bool value)
    {
        mChkBox->setChecked(value);
    }
private:
    QCheckBox *mChkBox;
    QString mExplanation;
};
*/

GpgOLConfigPage::GpgOLConfigPage(QWidget *parent):
    QWidget(parent)
{
    setupGUI();

    load();
}

/* Helper to build an "About" style layout.
static QLayout *buildAboutLayout(const QString &version)
{
    auto hLay = new QHBoxLayout;
    auto vLay = new QVBoxLayout;
    hLay->addLayout(vLay);
    hLay->addStretch(1);

    auto iconLbl = new QLabel;
    iconLbl->setPixmap(QIcon(":/gpgol-logo.png").pixmap(128, 80));
    auto versionLbl = new QLabel(QStringLiteral("    ") + QString::fromUtf8(_("Version ")) + version);
    vLay->addWidget(iconLbl);
    vLay->addWidget(versionLbl);

    return hLay;
}
*/

void GpgOLConfigPage::setupGUI()
{
    auto baseLay = new QVBoxLayout(this);

    mSMIMEGrp = new QGroupBox(_("Enable the S/MIME support"));
    mSMIMEGrp->setCheckable(true);
    mSMIMEGrp->setAlignment(Qt::AlignLeft);

    auto smimeLay = new QVBoxLayout(mSMIMEGrp);

    mPreferSMIMEChk = new QCheckBox(_("&Prefer S/MIME when automatically resolving recipients"));
    mPreferSMIMEChk->setToolTip(_("If automatic resolution is enabled, prefer S/MIME over OpenPGP if both are possible."));
    smimeLay->addWidget(mPreferSMIMEChk);

    baseLay->addWidget(mSMIMEGrp);

    // The general group
    auto generalGrp = new QGroupBox(_("General"));
    auto generalLay = new QVBoxLayout(generalGrp);
    generalGrp->setAlignment(Qt::AlignLeft);
    mAlwaysSigChk = new QCheckBox(_("&Sign new messages by default"));
    mAlwaysSigChk->setToolTip(_("Toggles the sign option for all new mails."));
    mAlwaysEncChk = new QCheckBox(_("&Encrypt new messages by default"));
    mAlwaysSigChk->setToolTip(_("Toggles the encrypt option for all new mails."));

    mReplyCryptChk = new QCheckBox(_("S&elect crypto settings automatically "
                                     "for reply and forward"));
    mReplyCryptChk->setToolTip(_("Toggles sign, encrypt options if the original mail was signed or encrypted."));

    mInlinePGPChk = new QCheckBox(_("&Send OpenPGP mails without attachments as PGP/Inline"));
    mInlinePGPChk->setToolTip(_("Instead of using the PGP/MIME format, "
                                "which properly handles attachments and encoding, "
                                "the deprecated PGP/Inline is used.\n"
                                "This can be useful for compatibility but should generally not "
                                "be used."));

    generalLay->addWidget(mAlwaysSigChk);
    generalLay->addWidget(mAlwaysEncChk);
    generalLay->addWidget(mReplyCryptChk);
    generalLay->addWidget(mInlinePGPChk);

    baseLay->addWidget(generalGrp);

    // The automation checkboxes
    mAutomationGrp = new QGroupBox(_("Automation"));
    mAutomationGrp->setToolTip(_("Enable or disable any automated key handling."));
    auto autoLayout = new QVBoxLayout(mAutomationGrp);
    mAutomationGrp->setCheckable(true);

    mAutoResolveChk = new QCheckBox(_("&Resolve recipient keys automatically"));
    autoLayout->addWidget(mAutoResolveChk);

    auto subLay = new QHBoxLayout;
    mAutoSecureChk = new QCheckBox(_("Automatically secure &messages"));
    mAutoSecureChk->setToolTip(_("Automatically toggles secure if keys with at least level 1 trust were found for all recipients."));
    subLay->addSpacing(20);
    subLay->addWidget(mAutoSecureChk);
    autoLayout->addLayout(subLay);

    mAutoTrustChk = new QCheckBox(QStringLiteral("%1 (%2)").arg(_("Include OpenPGP &trust based on communication history")).arg(_("experimental")));
    mAutoTrustChk->setToolTip(_("This changes the trust model to \"tofu+pgp\" which tracks the history of key usage. "
                                "Automated trust can <b>never</b> exceed level 2."));
    /* Dsiabled for now */
    mAutoTrustChk->setVisible(false);

    autoLayout->addWidget(mAutoTrustChk);

    baseLay->addWidget(mAutomationGrp);

    // The debugging group
    mDbgGrp = new QGroupBox(_("Enable Logging"));
    mDbgGrp->setCheckable(true);

    mDbgCombo = new QComboBox;
    mDbgCombo->addItem(_("Default"), 1);
    mDbgCombo->addItem(_("+Outlook API calls"), (DBG_OOM));
    mDbgCombo->addItem(_("+Memory analysis"), (DBG_OOM | DBG_MEMORY));
    mDbgCombo->addItem(_("+Call tracing"), (DBG_OOM | DBG_MEMORY | DBG_TRACE));

    mDbgVerboseWarningLabel = new QLabel(_("<b>Warning:</b> Decreased performance. Huge logs!"));

    mDbgComboLabel = new QLabel(_("Log level:"));

    mDbgLogFileLabel = new QLabel(_("Log File (required):"));

    mDbgDataChk = new QCheckBox(_("Include Mail contents (decrypted!) and meta information."));

    mDbgLogFileName = new QLineEdit;

    auto dbgLay = new QVBoxLayout(mDbgGrp);

    auto logFileLay = new QHBoxLayout;

    mDbgLogFileBtn = new QPushButton;
    mDbgLogFileBtn->setIcon(style()->standardIcon(QStyle::SP_FileDialogStart));

    logFileLay->addWidget(mDbgLogFileLabel, 0);
    logFileLay->addWidget(mDbgLogFileName, 1);
    logFileLay->addWidget(mDbgLogFileBtn, 0);
    dbgLay->addLayout(logFileLay);

    auto dbgComboLay = new QHBoxLayout;
    dbgLay->addLayout(dbgComboLay);

    dbgComboLay->addWidget(mDbgComboLabel);
    dbgComboLay->addWidget(mDbgCombo);
    dbgComboLay->addWidget(mDbgVerboseWarningLabel);
    dbgComboLay->addStretch(1);

    dbgLay->addWidget(mDbgDataChk);

    baseLay->addWidget(mDbgGrp);

    // End debugging group

    // baseLay->addLayout(buildAboutLayout(mVersion));

    baseLay->addStretch(1);

    connect(mAutoResolveChk, &QCheckBox::toggled, [this] (bool on) {
            mAutoSecureChk->setEnabled(on);
            mPreferSMIMEChk->setEnabled(mSMIMEGrp->isChecked() && on);
        });
    connect(mSMIMEGrp, &QGroupBox::toggled, [this] (bool on) {
            mPreferSMIMEChk->setEnabled(mAutoSecureChk->isChecked() && on);
        });
    connect(mDbgGrp, &QGroupBox::toggled, [this] (bool) {
            enableDisableDbgWidgets();
        });

    connect(mDbgCombo, &QComboBox::currentTextChanged, [this] (QString) {
            mDbgVerboseWarningLabel->setVisible((mDbgCombo->currentData().toInt() & DBG_TRACE));
        });

    connect(mDbgLogFileBtn, &QPushButton::clicked, [this] () {
        const auto fileName = QFileDialog::getSaveFileName(this, _("Select log file"),
                              mDbgLogFileName->text(),
                              "(*.txt)");
        if (!fileName.isEmpty()) {
            mDbgLogFileName->setText(QDir::toNativeSeparators(fileName));
        }
    });
    enableDisableDbgWidgets();
}

static bool strToBool(const std::string &str, bool defaultVal = false)
{
    if (str.empty()) {
      return defaultVal;
    }

    if (str == "1") {
      return true;
    }
    if (str == "0") {
      return false;
    }

    qDebug() << "Unknown bool val" << str.c_str();
    return defaultVal;
}

static bool loadBool(const char *name, bool defaultVal)
{
    return strToBool(W32::readRegStr(nullptr, GPGOL_REG_PATH, name), defaultVal);
}

/* Bump this if you remove a config value */
#define CONFIG_VERSION "1"

static const QMap<QString, bool> defaultMap {
    { QStringLiteral("enableSmime"), false },
    { QStringLiteral("encryptDefault"), false },
    { QStringLiteral("signDefault"), false },
    { QStringLiteral("inlinePGP"), false },
    { QStringLiteral("replyCrypt"), true },
    { QStringLiteral("preferSmime"), false },
    { QStringLiteral("debugGPGME"), false },
    { QStringLiteral("automation"), true },
    { QStringLiteral("autoresolve"), true },
    { QStringLiteral("autosecure"), true },
    { QStringLiteral("autotrust"), false },
    { QStringLiteral("automation"), true },
    { QStringLiteral("syncEnc"), false },
};

void GpgOLConfigPage::updateGUI(const QMap<QString, bool> &values)
{
    bool smimeEnabled = values["enableSmime"];
    mSMIMEGrp->setChecked(smimeEnabled);
    mPreferSMIMEChk->setChecked(values["preferSmime"]);

    mAlwaysEncChk->setChecked(values["encryptDefault"]);
    mAlwaysSigChk->setChecked(values["signDefault"]);
    mInlinePGPChk->setChecked(values["inlinePGP"]);
    mReplyCryptChk->setChecked(values["replyCrypt"]);

    mAutomationGrp->setChecked(values["automation"]);
    mAutoSecureChk->setChecked(values["autosecure"]);
    mAutoTrustChk->setChecked(values["autotrust"]);
    mAutoResolveChk->setChecked(values["autoresolve"]);

    mAutoSecureChk->setEnabled(mAutoResolveChk->isChecked() && mAutomationGrp->isChecked());
    mPreferSMIMEChk->setEnabled(mAutoResolveChk->isChecked() && smimeEnabled);
}

void GpgOLConfigPage::load()
{
    QMap<QString, bool> confValues;

    for (const auto &key: defaultMap.keys()) {
        confValues[key] = loadBool(key.toLocal8Bit().constData(), defaultMap[key]);
    }
    updateGUI(confValues);

    const std::string version = W32::readRegStr(nullptr, GPGOL_REG_PATH, "config-version");
    if (version != CONFIG_VERSION) {
        qDebug() << "Config update. Cleaning old values";
    }

    const auto logFile = W32::readRegStr(nullptr, GPGOL_REG_PATH, "logFile");
    mDbgLogFileName->setText(logFile.empty() ?
            QDir::toNativeSeparators(QStandardPaths::writableLocation(QStandardPaths::HomeLocation) + "/gpgol.txt") :
            QString::fromStdString(logFile));

    const auto logLevelS = W32::readRegStr(nullptr, GPGOL_REG_PATH, "enableDebug");

    bool ok;
    int logLevel = QString::fromStdString(logLevelS).toInt(&ok);
    if (!ok) {
        logLevel = 0;
    }

    mDbgGrp->setChecked(logLevel > 0);

    int idx = 0;
    if ((logLevel & DBG_OOM)) {
        idx++;
    }
    if ((logLevel & DBG_MEMORY)) {
        idx++;
    }
    if ((logLevel & DBG_TRACE)) {
        idx++;
    }

    mDbgCombo->setCurrentIndex(idx);

    mDbgDataChk->setChecked((logLevel & DBG_DATA));
}

void GpgOLConfigPage::defaults()
{
    updateGUI(defaultMap);
}

static void saveBool(const char *name, bool value)
{
    const char *val = value ? "1" : "0";

    if (!W32::writeRegStr(nullptr, GPGOL_REG_PATH, name, val)) {
        qWarning() << "Failed to write registry value for" << name;
    }
}

static void saveInt(const char *name, int value)
{
    const std::string val = std::to_string(value);

    if (!W32::writeRegStr(nullptr, GPGOL_REG_PATH, name, val.c_str())) {
        qWarning() << "Failed to write registry value for" << name;
    }
}

void GpgOLConfigPage::save() const
{
    saveBool("enableSmime", mSMIMEGrp->isChecked());
    saveBool("preferSmime", mPreferSMIMEChk->isChecked());

    saveBool("encryptDefault", mAlwaysEncChk->isChecked());
    saveBool("signDefault", mAlwaysSigChk->isChecked());
    saveBool("inlinePGP", mInlinePGPChk->isChecked());
    saveBool("replyCrypt", mReplyCryptChk->isChecked());

    saveBool("automation", mAutomationGrp->isChecked());
    saveBool("autosecure", mAutoSecureChk->isChecked());
    saveBool("autotrust", mAutoTrustChk->isChecked());
    saveBool("autoresolve", mAutoResolveChk->isChecked());

    int logLevel = mDbgCombo->currentData().toInt();
    logLevel |= mDbgDataChk->isChecked() ? DBG_DATA : 0;

    saveInt("enableDebug", logLevel);
    W32::writeRegStr(nullptr, GPGOL_REG_PATH, "logFile", QDir::toNativeSeparators(
                mDbgLogFileName->text()).toLocal8Bit().constData());

    W32::writeRegStr(nullptr, GPGOL_REG_PATH, "config-version", CONFIG_VERSION);
}

void GpgOLConfigPage::enableDisableDbgWidgets()
{
    bool vis = mDbgGrp->isChecked();

    mDbgDataChk->setVisible(vis);
    mDbgCombo->setVisible(vis);
    mDbgComboLabel->setVisible(vis);

    mDbgLogFileName->setVisible(vis);
    mDbgLogFileLabel->setVisible(vis);
    mDbgLogFileBtn->setVisible(vis);

    mDbgVerboseWarningLabel->setVisible(vis && (mDbgCombo->currentData().toInt() & DBG_TRACE));
}

#include "gpgolconfigpage.moc"
